mod vm;

use std::sync::Arc;
use vm::FootVM;

use pixels::{Pixels, SurfaceTexture};

use winit::application::ApplicationHandler;
use winit::dpi::LogicalSize;
use winit::event::{StartCause, WindowEvent};
use winit::event_loop::{ActiveEventLoop, ControlFlow, EventLoop};
use winit::window::{Window, WindowAttributes, WindowButtons, WindowId};

pub struct App
{
    vm: FootVM,
    window: Option<Arc<Window>>,
    pixels: Option<Pixels<'static>>
}

impl App
{
    fn new(vm: FootVM) -> Self
    {
        App
        {
            vm,
            window: None,
            pixels: None
        }
    }

    fn update(&mut self)
    {
        self.vm.cycle();
    }

    fn draw(&mut self)
    {
        let pixels = &mut self.pixels.as_mut().unwrap();
        let frame = pixels.frame_mut();

        for i in 0u16..32u16
        {
            for j in 0u16..32u16
            {
                let index = i + j * 32;
                let mut value = self.vm[index + 0xFBFF];

                for k in 0..16
                {
                    let pixel_value = ((value & 1) * 255) as u8;
                    let pixel_index = (index as usize * 64) + k * 4;
                    frame[pixel_index + 0] = pixel_value;
                    frame[pixel_index + 1] = pixel_value;
                    frame[pixel_index + 2] = pixel_value;
                    frame[pixel_index + 3] = 0xFF;

                    value >>= 1;
                }
            }
        }

        pixels.render().expect("");
    }
}

impl ApplicationHandler for App
{
    fn new_events(&mut self, _: &ActiveEventLoop, cause: StartCause)
    {
        match cause
        {
            StartCause::Poll =>
                {
                    self.update();
                    self.draw()
                },
            _ => {}
        }
    }

    fn resumed(&mut self, event_loop: &ActiveEventLoop)
    {
        const WIDTH: u32 = 128;
        const HEIGHT: u32 = 128;

        let window = event_loop.create_window(
            WindowAttributes::default()
                .with_resizable(false)
                .with_enabled_buttons(WindowButtons::CLOSE | WindowButtons::MINIMIZE)
                .with_inner_size(LogicalSize::new(WIDTH * 4, HEIGHT * 4))
                .with_title("foot vm")
        ).unwrap();
        let window = Arc::new(window);

        self.window = Some(window);

        let surface_texture = SurfaceTexture::new(WIDTH * 4, HEIGHT * 4, self.window.as_ref().unwrap().clone());
        self.pixels = Some(Pixels::new(WIDTH, HEIGHT, surface_texture).unwrap());
    }

    fn window_event(&mut self, event_loop: &ActiveEventLoop, _: WindowId, event: WindowEvent)
    {
        match event
        {
            WindowEvent::RedrawRequested => self.draw(),
            WindowEvent::CloseRequested => event_loop.exit(),
            _ => {}
        }
    }
}

const ASM_SRC: &'static str = include_str!("../examples/first.asm");

fn main()
{
    let mut vm = FootVM::new();
    vm.load(ASM_SRC, Some(&mut std::io::stderr())).expect("");

    let event_loop = EventLoop::new().unwrap();
    event_loop.set_control_flow(ControlFlow::Poll);

    let mut app = App::new(vm);
    event_loop.run_app(&mut app).expect("");
}
